from functools import wraps
from azure.cosmos.aio import CosmosClient
import ramda as R
from .. import KEY, ENDPOINT

def with_client(coro):
    """Runs `coro` with client (if provided), otherwise runs within a new client's async context
    - `async def coro(*args, *, client: CosmosClient)`
    """
    @wraps(coro)
    async def f(
        *args, client: CosmosClient | None = None,
        url: str | None = ENDPOINT, key: str | None = KEY, **kwargs
    ):
        kwargs = R.omit(["client", "url", "key"], kwargs)
        if client is None:
            assert url is not None and key is not None, "Provide a key+endpoint or a client"
            async with CosmosClient(url=url, credential=key) as client:
                return await coro(*args, client=client, **kwargs)
        else:
            return await coro(*args, client=client, **kwargs)
    return f

def client(url: str = ENDPOINT, key: str = KEY) -> CosmosClient:
    return CosmosClient(url=url, credential=key)


"""
### Cosmos DB
> Python SDK for Azure Cosmos DB NoSQL

- `.env`: by default, `COSMOS_ENDPOINT` and `COSMOS_KEY` are retrieved from environment vars. If unset, you'll have to provide or set them manually
"""
from .env import ENDPOINT, KEY
from .util import client, CosmosClient
from . import util, list, create, delete, query, patch


from typing import Literal, Annotated, Any
from pydantic import BaseModel, StringConstraints, model_serializer
from azure.cosmos.aio import CosmosClient
import ramda as R
from .. import KEY, ENDPOINT
from ..util import with_client

Path = Annotated[str, StringConstraints(pattern='/.*')]

class ValueOp(BaseModel):
    op: Literal['add', 'replace', 'set', 'incr', 'remove']
    value: Any
    path: Path

class MoveOp(BaseModel):
    op: Literal['move'] = 'move'
    from_: Path
    to: Path
    
    @model_serializer
    def dump(self) -> dict[str, Any]:
        return {
            'op': self.op,
            'from': self.from_,
            'path': self.to
        }

Op = ValueOp | MoveOp

@with_client
async def item(
    database: str, container: str, id: str, ops: list[Op], partition_key: str | None = None,
    *, client: CosmosClient | None = None, url: str | None = ENDPOINT, key: str | None = KEY,
) -> dict[str]:
    """Partial update (PATCH) a given item. See [Azure's docs](https://learn.microsoft.com/en-us/azure/cosmos-db/partial-document-update) and [python SDK docs](https://learn.microsoft.com/en-us/azure/cosmos-db/partial-document-update-getting-started?tabs=python) for details
    - Returns the patched object"""
    cc = client.get_database_client(database).get_container_client(container)
    dict_ops = [op.model_dump(mode='json') for op in ops]
    return await cc.patch_item(item=id, partition_key=partition_key or id, patch_operations=dict_ops)



from azure.cosmos.aio import CosmosClient
from azure.cosmos import PartitionKey
from .. import KEY, ENDPOINT
from ..util import with_client

@with_client
async def db(
    database: str, *, client: CosmosClient | None = None,
    url: str = ENDPOINT, key: str = KEY, **kwargs
):
    """Creates a new DB"""
    await client.create_database(database, **kwargs)

@with_client
async def container(
    database: str, container: str, partition_key: str = "/id", *,
    client: CosmosClient | None = None, url: str | None = ENDPOINT, key: str | None = KEY,
    **kwargs
):
    """Creates container within DB (also creates database if needed)"""
    db = await client.create_database_if_not_exists(database)
    key_path = PartitionKey(path=partition_key)
    await db.create_container(id=container, partition_key=key_path, **kwargs)
    
@with_client
async def item(
    database: str, container: str, body: dict[str], replace: bool = True,
    *, client: CosmosClient | None = None, url: str | None = ENDPOINT, key: str | None = KEY,
) -> dict[str]:
    db = client.get_database_client(database)
    cc = db.get_container_client(container)
    if replace:
        return await cc.upsert_item(body)
    else:
        return await cc.create_item(body)



from typing import Annotated
from pydantic import BaseModel
from azure.cosmos.aio import CosmosClient
import ramda as R
import haskellian.asyn as hka
from .. import KEY, ENDPOINT
from ..util import with_client

@with_client
async def read(
    database: str, container: str,
    id: int | str, partition_key: str | None = None,
    *, client: CosmosClient, url: str = ENDPOINT, key: str = KEY
) -> dict[str]:
    """Point read of an item given its `id` and `partition_key`
    - If `partition_key is None`, it'll be set to `id`
    """
    partition_key = partition_key or id
    cc = client.get_database_client(database).get_container_client(container)
    return await cc.read_item(str(id), partition_key)
    
def name_val(name: str):
    assert name.startswith('@'), "name must start with '@'"
    return name
    
class Param(BaseModel):
    name: Annotated[str, name_val]
    value: str
    
    
@with_client
async def sql(
    database: str, container: str,
    query: str, params: list[Param] = [], partition_key: str | None = None,
    *, client: CosmosClient, url: str = ENDPOINT, key: str = KEY
) -> list[dict[str]]:
    """
    Query via CosmosDB SQL query
    - `query`: e.g "SELECT * FROM x WHERE x.id = @value" (`FROM _` can be anything, e.g. `FROM me`)
    - `params`: e.g. [Param(name="@value", value=user_id)]
    """
    cont = client.get_database_client(database).get_container_client(container)
    return await hka.syncify(cont.query_items(
        query=query, parameters=R.map(Param.model_dump, params),
        partition_key=partition_key
    ))


from azure.cosmos.aio import CosmosClient
import haskellian.asynch as hka
from .. import KEY, ENDPOINT
from ..util import with_client

@with_client
async def dbs(
    *, client: CosmosClient | None, url: str | None = ENDPOINT, key: str | None = KEY
) -> list[dict[str]]:
    """List all databases of the Cosmos DB resource"""
    return await hka.syncify(client.list_databases())
    
@with_client
async def containers(
    database: str, *, client: CosmosClient | None, url: str = ENDPOINT, key: str = KEY
) -> list[dict[str]]:
    """List all containers of a given database"""
    db = client.get_database_client(database)
    return await hka.syncify(db.list_containers())

@with_client
async def items(
    database: str, container: str, *,
    client: CosmosClient | None, url: str = ENDPOINT, key: str = KEY
) -> list[dict[str]]:
    """List all items of a given container (within a database)"""
    db = client.get_database_client(database)
    cc = db.get_container_client(container)
    return await hka.syncify(cc.read_all_items())



from azure.cosmos.aio import CosmosClient
from .. import KEY, ENDPOINT
from ..util import with_client

@with_client
async def db(
    database: str, *, client: CosmosClient, url: str = ENDPOINT, key: str = KEY
):
    await client.delete_database(database)

@with_client     
async def container(
    database: str, container: str, *, client: CosmosClient, url: str = ENDPOINT, key: str = KEY
):
    db = client.get_database_client(database)
    await db.delete_container(container)

@with_client
async def item(
    database: str, container: str, id: str, partition_key: str | None = None,
    *, client: CosmosClient, url: str | None = ENDPOINT, key: str | None = KEY
):
    """Delete an item given its `id` and `partition_key`
    - If `partition_key is None`, it'll be set to `id`
    """
    partition_key = partition_key or id
    db = client.get_database_client(database)
    cc = db.get_container_client(container)
    await cc.delete_item(id, partition_key)
